/* SPDX-License-Identifier: BSD-2-Clause */
/*******************************************************************************
 * Copyright 2018-2019, Fraunhofer SIT sponsored by Infineon Technologies AG
 * All rights reserved.
 ******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h" // IWYU pragma: keep
#endif

#include <stdint.h> // for uint8_t
#include <stdlib.h> // for malloc, size_t, NULL
#include <string.h> // for memcpy

#include "fapi_int.h"       // for FAPI_CONTEXT, PATH_GET_DESCRIPTION_READ
#include "fapi_types.h"     // for UINT8_ARY
#include "fapi_util.h"      // for ifapi_session_init
#include "ifapi_io.h"       // for ifapi_io_poll
#include "ifapi_keystore.h" // for ifapi_cleanup_ifapi_object, IFAPI_OBJECT
#include "ifapi_macros.h"   // for check_not_null, return_if_error_reset_state
#include "tss2_common.h"    // for TSS2_RC, TSS2_RC_SUCCESS, TSS2_BASE_RC_T...
#include "tss2_fapi.h"      // for FAPI_CONTEXT, Fapi_GetAppData, Fapi_GetA...

#define LOGMODULE fapi
#include "util/log.h" // for LOG_TRACE, return_if_error, base_rc, got...

/** One-Call function for Fapi_GetAppData
 *
 * Every object has a description field that can be retrieved in order to obtain
 * additional information in its "path" entry.
 *
 * @param[in,out] context The FAPI_CONTEXT
 * @param[in] path The path to the object for which the appData is returned
 * @param[out] appData A copy of the appData. May be NULL (callee-allocated)
 * @param[out] appDataSize The size of the returned AppData. May be NULL
 *
 * @retval TSS2_RC_SUCCESS: if the function call was a success.
 * @retval TSS2_FAPI_RC_BAD_REFERENCE: if context or path is NULL.
 * @retval TSS2_FAPI_RC_BAD_CONTEXT: if context corruption is detected.
 * @retval TSS2_FAPI_RC_BAD_PATH: if path does not map to a FAPI entity.
 * @retval TSS2_FAPI_RC_BAD_SEQUENCE: if the context has an asynchronous
 *         operation already pending.
 * @retval TSS2_FAPI_RC_IO_ERROR: if the data cannot be saved.
 * @retval TSS2_FAPI_RC_MEMORY: if the FAPI cannot allocate enough memory for
 *         internal operations or return parameters.
 * @retval TSS2_FAPI_RC_NO_TPM if FAPI was initialized in no-TPM-mode via its
 *         config file.
 * @retval TSS2_FAPI_RC_PATH_NOT_FOUND if a FAPI object path was not found
 *         during authorization.
 * @retval TSS2_FAPI_RC_KEY_NOT_FOUND if a key was not found.
 * @retval TSS2_FAPI_RC_BAD_VALUE if an invalid value was passed into
 *         the function.
 * @retval TSS2_FAPI_RC_TRY_AGAIN if an I/O operation is not finished yet and
 *         this function needs to be called again.
 * @retval TSS2_FAPI_RC_GENERAL_FAILURE if an internal error occurred.
 * @retval TSS2_FAPI_RC_NOT_PROVISIONED FAPI was not provisioned.
 */
TSS2_RC
Fapi_GetAppData(FAPI_CONTEXT *context, char const *path, uint8_t **appData, size_t *appDataSize) {
    LOG_TRACE("called for context:%p", context);

    TSS2_RC r;

    /* Check for NULL parameters */
    check_not_null(context);
    check_not_null(path);

    r = Fapi_GetAppData_Async(context, path);
    return_if_error_reset_state(r, "Path_SetDescription");

    do {
        /* We wait for file I/O to be ready if the FAPI state automata
           are in a file I/O state. */
        r = ifapi_io_poll(&context->io);
        return_if_error(r, "Something went wrong with IO polling");

        /* Repeatedly call the finish function, until FAPI has transitioned
           through all execution stages / states of this invocation. */
        r = Fapi_GetAppData_Finish(context, appData, appDataSize);
    } while (base_rc(r) == TSS2_BASE_RC_TRY_AGAIN);

    return_if_error_reset_state(r, "Path_SetDescription");

    LOG_TRACE("finished");
    return TSS2_RC_SUCCESS;
}

/** Asynchronous function for Fapi_GetAppData
 *
 * Every object has a description field that can be retrieved in order to obtain
 * additional information in its "path" entry.
 *
 * Call Fapi_GetAppData_Finish to finish the execution of this command.
 *
 * @param[in,out] context The FAPI_CONTEXT
 * @param[in] path The path to the object for which the appData is returned
 *
 * @retval TSS2_RC_SUCCESS: if the function call was a success.
 * @retval TSS2_FAPI_RC_BAD_REFERENCE: if context or path is NULL.
 * @retval TSS2_FAPI_RC_BAD_CONTEXT: if context corruption is detected.
 * @retval TSS2_FAPI_RC_BAD_PATH: if path does not map to a FAPI entity.
 * @retval TSS2_FAPI_RC_BAD_SEQUENCE: if the context has an asynchronous
 *         operation already pending.
 * @retval TSS2_FAPI_RC_IO_ERROR: if the data cannot be saved.
 * @retval TSS2_FAPI_RC_MEMORY: if the FAPI cannot allocate enough memory for
 *         internal operations or return parameters.
 * @retval TSS2_FAPI_RC_NO_TPM if FAPI was initialized in no-TPM-mode via its
 *         config file.
 * @retval TSS2_FAPI_RC_PATH_NOT_FOUND if a FAPI object path was not found
 *         during authorization.
 * @retval TSS2_FAPI_RC_KEY_NOT_FOUND if a key was not found.
 * @retval TSS2_FAPI_RC_BAD_VALUE if an invalid value was passed into
 *         the function.
 * @retval TSS2_FAPI_RC_NOT_PROVISIONED FAPI was not provisioned.
 */
TSS2_RC
Fapi_GetAppData_Async(FAPI_CONTEXT *context, char const *path) {
    LOG_TRACE("called for context:%p", context);
    LOG_TRACE("path: %s", path);

    TSS2_RC r;

    /* Check for NULL parameters */
    check_not_null(context);
    check_not_null(path);

    /* Cleanup command context. */
    memset(&context->cmd, 0, sizeof(IFAPI_CMD_STATE));

    /* Reset all context-internal session state information. */
    r = ifapi_session_init(context);
    return_if_error(r, "Initialize GetAppData");

    /* Load the object metadata from keystore. */
    r = ifapi_keystore_load_async(&context->keystore, &context->io, path);
    return_if_error2(r, "Could not open: %s", path);

    context->state = PATH_GET_DESCRIPTION_READ;
    LOG_TRACE("finished");
    return TSS2_RC_SUCCESS;
}

/** Asynchronous finish function for Fapi_GetAppData
 *
 * This function should be called after a previous Fapi_GetAppData_Async.
 *
 * @param[in,out] context The FAPI_CONTEXT
 * @param[out] appData A copy of the appData. May be Null (callee-allocated)
 * @param[out] appDataSize The size of the returned AppData. May be NULL
 *
 * @retval TSS2_RC_SUCCESS: if the function call was a success.
 * @retval TSS2_FAPI_RC_BAD_REFERENCE: if context is NULL.
 * @retval TSS2_FAPI_RC_BAD_CONTEXT: if context corruption is detected.
 * @retval TSS2_FAPI_RC_BAD_SEQUENCE: if the context has an asynchronous
 *         operation already pending.
 * @retval TSS2_FAPI_RC_IO_ERROR: if the data cannot be saved.
 * @retval TSS2_FAPI_RC_MEMORY: if the FAPI cannot allocate enough memory for
 *         internal operations or return parameters.
 * @retval TSS2_FAPI_RC_TRY_AGAIN: if the asynchronous operation is not yet
 *         complete. Call this function again later.
 * @retval TSS2_FAPI_RC_BAD_PATH if a path is used in inappropriate context
 *         or contains illegal characters.
 * @retval TSS2_FAPI_RC_GENERAL_FAILURE if an internal error occurred.
 * @retval TSS2_FAPI_RC_BAD_VALUE if an invalid value was passed into
 *         the function.
 */
TSS2_RC
Fapi_GetAppData_Finish(FAPI_CONTEXT *context, uint8_t **appData, size_t *appDataSize) {
    LOG_TRACE("called for context:%p", context);

    TSS2_RC      r;
    IFAPI_OBJECT object;
    UINT8_ARY   *objAppData;

    /* Check for NULL parameters */
    check_not_null(context);

    switch (context->state) {
    statecase(context->state, PATH_GET_DESCRIPTION_READ);
        r = ifapi_keystore_load_finish(&context->keystore, &context->io, &object);
        return_try_again(r);
        return_if_error_reset_state(r, "read_finish failed");

        /* Get the application data from the metadata objects. */
        switch (object.objectType) {
        case IFAPI_KEY_OBJ:
            objAppData = &object.misc.key.appData;
            break;
        case IFAPI_NV_OBJ:
            objAppData = &object.misc.nv.appData;
            break;
        default:
            goto_error(r, TSS2_FAPI_RC_BAD_PATH, "Object has no app data.", cleanup);
        }

        if (appData) {
            /* Duplicate the application data to be returned to the caller. */
            if (objAppData->size) {
                *appData = malloc(objAppData->size);
                goto_if_null2(*appData, "Out of memory.", r, TSS2_FAPI_RC_MEMORY, cleanup);
                memcpy(*appData, &objAppData->buffer[0], objAppData->size);
            } else {
                *appData = NULL;
            }
        }
        if (appDataSize)
            *appDataSize = objAppData->size;

        r = TSS2_RC_SUCCESS;
        break;

    statecasedefault(context->state);
    }

cleanup:
    /* Cleanup any intermediate results and state stored in the context. */
    ifapi_cleanup_ifapi_object(&object);
    ifapi_cleanup_ifapi_object(&context->loadKey.auth_object);
    ifapi_cleanup_ifapi_object(context->loadKey.key_object);
    ifapi_cleanup_ifapi_object(&context->createPrimary.pkey_object);
    context->state = FAPI_STATE_INIT;
    LOG_TRACE("finished");
    return r;
}
